import logging
from pathlib import Path

from pepyno.constants import NAME


def setup_logging(log_level: int = logging.WARNING, log_dir: str = "./") -> logging.Logger:
    """Set up and configure logging with both file and console handlers.

    Args:
        log_level: Logging level to use
        log_dir: Directory to store log files

    Returns:
        Configured logger instance
    """
    log_path = Path(log_dir)
    log_path.mkdir(exist_ok=True)

    log_formatter = logging.Formatter(
        "%(asctime)s [%(threadName)-12.12s] [%(levelname)-10.10s]  %(message)s"
    )
    short_formatter = logging.Formatter("[%(levelname)-8.8s]  %(message)s")

    log = logging.getLogger(NAME)
    log.setLevel(log_level)

    if log.handlers:
        log.handlers.clear()

    # File handler
    file_path = log_path / f"{NAME}.log"
    file_handler = logging.FileHandler(file_path)
    file_handler.setFormatter(log_formatter)
    log.addHandler(file_handler)

    # Console handler
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(short_formatter)
    log.addHandler(console_handler)

    return log



import json
from pathlib import Path

import pytest

from src.pepyno.converter import convert

FIXTURES_DIR = Path(__file__).parent / "fixtures"
BEHAVE_JSON = FIXTURES_DIR / "behave.json"
EXPECTED_JSON = FIXTURES_DIR / "expected.json"
AUTORETRY_BEHAVE_JSON = FIXTURES_DIR / "autoretry.json"
AUTORETRY_EXPECTED_JSON = FIXTURES_DIR / "autoretry_expected.json"
AUTORETRY_DEDUPE_JSON = FIXTURES_DIR / "autoretry_dedupe.json"


def test_convert():
    """Test basic conversion functionality."""
    with open(BEHAVE_JSON, encoding="utf-8") as f:
        converted = convert(json.load(f))
    with open(EXPECTED_JSON, encoding="utf-8") as f:
        expected_result = json.load(f)
    assert sorted(converted) == sorted(expected_result)


def test_autoretry_convert():
    """Test conversion with auto-retry features."""
    with open(AUTORETRY_BEHAVE_JSON, encoding="utf-8") as f:
        converted = convert(json.load(f))
    with open(AUTORETRY_EXPECTED_JSON, encoding="utf-8") as f:
        expected_result = json.load(f)
    assert sorted(converted) == sorted(expected_result)


def test_dedupe_convert():
    """Test conversion with deduplication enabled."""
    with open(AUTORETRY_BEHAVE_JSON, encoding="utf-8") as f:
        converted = convert(json.load(f), deduplicate=True)
    with open(AUTORETRY_DEDUPE_JSON, encoding="utf-8") as f:
        expected_result = json.load(f)
    assert sorted(converted) == sorted(expected_result)


def test_ids_are_unique():
    """Test that all IDs in the converted output are unique."""
    with open(BEHAVE_JSON, encoding="utf-8") as f:
        converted = convert(json.load(f))
        ids = []
        for feature in converted:
            ids.append(feature["id"])
            for element in feature["elements"]:
                ids.append(element["id"])
    assert len(set(ids)) == 5


@pytest.fixture
def behave_json():
    """Fixture that provides the behave JSON data."""
    with open(BEHAVE_JSON, encoding="utf8") as f:
        return json.load(f)


@pytest.fixture
def autoretry_json():
    """Fixture that provides the autoretry JSON data."""
    with open(AUTORETRY_BEHAVE_JSON, encoding="utf-8") as f:
        return json.load(f)


# Example of a parametrized test using fixtures
@pytest.mark.parametrize("deduplicate", [False, True])
def test_parametrized_convert(behave_json, deduplicate):
    """Example of a parametrized test demonstrating both with and without deduplication."""
    converted = convert(behave_json, deduplicate=deduplicate)
    assert isinstance(converted, list)
    assert len(converted) > 0



import json
import time
from datetime import datetime
from logging import Logger
from pathlib import Path

from pepyno.constants import NAME, VERSION
from pepyno.converter import convert


def process_file(
    log: Logger,
    infile: Path,
    outfile: Path | None = None,
    remove_background: bool = False,
    duration_format: bool = False,
    deduplicate: bool = False,
    pretty: bool = False,
) -> list:
    """Process the input file and generate Cucumber-compatible output.

    Args:
        log: a Logger instance
        infile: Input JSON file path
        outfile: Output JSON file path (optional)
        remove_background: Whether to remove background steps
        duration_format: Whether to format duration values
        deduplicate: Whether to remove duplicate scenarios
        pretty: Whether to pretty-print the JSON output

    Returns:
        Processed JSON data
    """
    log.info(f"Processing file: {infile}")

    try:
        with open(infile, encoding="utf-8") as f:
            try:
                input_data = json.load(f)
            except json.JSONDecodeError as e:
                raise json.JSONDecodeError(f"Invalid JSON in {infile}: {str(e)}", e.doc, e.pos)

        # Process the data
        start_time = time.time()
        cucumber_output = convert(
            input_data,
            remove_background=remove_background,
            duration_format=duration_format,
            deduplicate=deduplicate,
        )
        log.info(f"Conversion completed in {time.time() - start_time:.2f} seconds")

        # Add metadata to the output
        if isinstance(cucumber_output, list) and cucumber_output:
            metadata = {
                "generated_by": f"{NAME} v{VERSION}",
                "timestamp": datetime.now().isoformat(),
                "source_file": str(infile),
            }
            for feature in cucumber_output:
                feature["metadata"] = metadata

        # Output the result
        if outfile:
            outfile.parent.mkdir(parents=True, exist_ok=True)
            with open(outfile, "w", encoding="utf-8") as f:
                indent = 4 if pretty else None
                json.dump(cucumber_output, f, indent=indent, ensure_ascii=False)
            log.info(f"Output written to: {outfile}")

        return cucumber_output

    except Exception as e:
        log.exception(f"Error processing file: {str(e)}")
        raise



def convert(json_file, remove_background=False, duration_format=False, deduplicate=False):
    """
    Convert Behave JSON format to Cucumber JSON format.

    Args:
        json_file: The JSON data from Behave to be converted
        remove_background: Whether to remove background elements
        duration_format: Whether to format duration values
        deduplicate: Whether to remove duplicate scenarios

    Returns:
        The converted JSON data in Cucumber format
    """
    # json_nodes are the scopes available in behave/cucumber json:
    # Feature -> elements(Scenarios) -> Steps
    json_nodes = ["feature", "elements", "steps"]
    # These fields don't exist in cucumber report, therefore when
    # converting from behave, we need to delete them
    fields_not_exist_in_cucumber_json = ["status", "step_type"]

    def format_level(tree, index=0, id_counter=0):
        for item in tree:
            # Location in behave json translates to uri and line in cucumber json
            uri, line_number = item.pop("location").split(":")
            item["line"] = int(line_number)

            # Remove fields that don't exist in cucumber json
            for field in fields_not_exist_in_cucumber_json:
                item.pop(field, None)  # Using pop with default to avoid KeyError

            if "tags" in item:
                print(f"Before: {item['tags']}")
                item["tags"] = [
                    {"name": "@" + tag if tag.startswith("auto") else tag, "line": item["line"] - 1}
                    for tag in item["tags"]
                ]
                print(f"After: {item['tags']}")

            if json_nodes[index] == "steps":
                if "result" in item:
                    # Truncate long error messages to maximum 2000 chars
                    if "error_message" in item["result"]:
                        error_msg = item["result"].pop("error_message")
                        item["result"]["error_message"] = str(
                            str(error_msg).replace('"', "").replace("\\'", "")
                        )[:2000]

                    # Format duration if needed
                    if "duration" in item["result"] and duration_format:
                        item["result"]["duration"] = int(item["result"]["duration"] * 1_000_000_000)
                else:
                    # In behave, skipped tests don't have result object in their json
                    # For every skipped test we need to generate a new result with status skipped
                    item["result"] = {"status": "skipped", "duration": 0}

                # Convert table format
                if "table" in item:
                    item["rows"] = []
                    t_line = 1
                    item["rows"].append(
                        {"cells": item["table"]["headings"], "line": item["line"] + t_line}
                    )
                    for table_row in item["table"]["rows"]:
                        t_line += 1
                        item["rows"].append({"cells": table_row, "line": item["line"] + t_line})
            else:
                # uri is the name of the feature file the current item is located in
                item["uri"] = uri
                item["description"] = ""
                item["id"] = id_counter
                id_counter += 1

            # If the scope is not "steps" proceed with the recursion
            if index != 2 and json_nodes[index + 1] in item:
                item[json_nodes[index + 1]] = format_level(
                    item[json_nodes[index + 1]], index + 1, id_counter=id_counter
                )
        return tree

    # Option to remove background element because behave pushes its steps to all scenarios already
    if remove_background:
        for feature in json_file:
            if feature["elements"] and feature["elements"][0].get("type") == "background":
                feature["elements"].pop(0)

    if deduplicate:

        def check_dupe(current_feature, current_scenario, previous_scenario):
            """Check if the current scenario is a duplicate of the previous one."""
            # Skip deduplication if not marked with autoretry tag
            if not any(
                "autoretry" in (tag.get("name", "") if isinstance(tag, dict) else tag)
                for tags in [current_feature.get("tags", []), current_scenario.get("tags", [])]
                for tag in tags
            ):
                return False

            # Compare scenario attributes
            for attr in ["keyword", "location", "name", "tags", "type"]:
                if previous_scenario.get(attr) != current_scenario.get(attr):
                    return False

            return True

        for feature in json_file:
            # Create a working list
            scenarios = []

            # For each scenario in the feature
            for scenario in feature.get("elements", []):
                # Append the scenario to the working list
                scenarios.append(scenario)

                # Check if this is a duplicate of the previous scenario
                if len(scenarios) >= 2:
                    previous_scenario = scenarios[-2]
                    if check_dupe(feature, scenario, previous_scenario):
                        # Remove the earlier scenario from the working list
                        scenarios.pop(-2)

            # Replace the existing list with the working list
            feature["elements"] = scenarios

    # Begin the recursion
    return format_level(json_file)



import json
from pathlib import PosixPath

import pytest

from pepyno.cli import main


@pytest.fixture
def sample_output_file(tmp_path):
    return tmp_path / "output.json"


def test_main_no_args(runner):
    result = runner.invoke(main, [])
    assert result.exit_code != 0
    assert "Error: Missing option '--infile' / '-i'" in result.output


def test_main_with_input_file(runner, sample_input_file):
    result = runner.invoke(main, ["-i", str(sample_input_file)])
    assert result.exit_code == 0
    assert "Test Scenario" in result.output


def test_main_with_output_file(runner, sample_input_file, sample_output_file):
    result = runner.invoke(main, ["-i", str(sample_input_file), "-o", str(sample_output_file)])
    assert result.exit_code == 0
    assert sample_output_file.exists()
    with sample_output_file.open() as f:
        data = json.load(f)
        assert data[0]["name"] == "Test Scenario"


def test_main_with_debug_flag(mocker, runner, sample_input_file):
    log_mock = mocker.patch("pepyno.cli.setup_logging")

    result = runner.invoke(main, ["-i", str(sample_input_file), "-d", 5])
    assert result.exit_code == 0
    log_mock.assert_called_once_with(0, PosixPath("./"))


def test_main_keyboard_interrupt(mocker, runner, sample_input_file):
    mocker.patch("pepyno.cli.process_file", side_effect=KeyboardInterrupt)
    result = runner.invoke(main, ["-i", str(sample_input_file)])
    assert result.exit_code == 130


def test_main_unhandled_exception(mocker, runner, sample_input_file):
    mocker.patch("pepyno.cli.process_file", side_effect=Exception("Unexpected Error"))
    result = runner.invoke(main, ["-i", str(sample_input_file)])
    assert result.exit_code == 1



import json
import logging
import pathlib
import sys
from pprint import pprint

import click

from pepyno.constants import NAME, VERSION
from pepyno.logger import setup_logging
from pepyno.tools import process_file


@click.command(context_settings={"help_option_names": ["-h", "--help"]})
@click.version_option(version=f"v{VERSION} 🥒", package_name=NAME)
@click.option(
    "--debug",
    "-d",
    type=click.IntRange(0, 5),
    default=0,
    help="Set debug level (0-5, where 5 is most verbose)",
)
@click.option(
    "--infile",
    "-i",
    type=click.Path(exists=True, path_type=pathlib.Path),
    required=True,
    help="Specify the input JSON file",
)
@click.option(
    "--outfile",
    "-o",
    type=click.Path(path_type=pathlib.Path),
    help="Specify the output JSON file, otherwise use stdout",
)
@click.option(
    "--remove-background",
    "-r",
    is_flag=True,
    help="Remove background steps from output",
)
@click.option("--format-duration", "-f", is_flag=True, help="Format the duration values")
@click.option(
    "--deduplicate",
    "-D",
    is_flag=True,
    help="Remove duplicate scenarios caused by @autoretry",
)
@click.option("--pretty", is_flag=True, help="Pretty-print the JSON output")
@click.option(
    "--log-dir",
    type=click.Path(file_okay=False, path_type=pathlib.Path),
    default="./",
    help="Directory to store log files",
)
def main(
    debug,
    infile,
    outfile,
    remove_background,
    format_duration,
    deduplicate,
    pretty,
    log_dir,
):
    log_level = 50 - (debug * 10)
    log = setup_logging(log_level, log_dir)

    log.info("Log level set to: %s", logging.getLevelName(log_level))
    log.debug("Arguments received: %s", locals())

    try:
        result = process_file(
            log,
            infile,
            outfile,
            remove_background,
            format_duration,
            deduplicate,
            pretty,
        )

        if not outfile:
            if pretty:
                print(json.dumps(result, indent=4, ensure_ascii=False))
            else:
                pprint(result)

    except KeyboardInterrupt:
        log.warning("Operation interrupted by user")
        sys.exit(130)
    except Exception as e:
        log.critical("Unhandled error: %s", e, exc_info=True)
        sys.exit(1)


if __name__ == "__main__":
    main()
